import pytest

from sokoenginepy.game import (
    BaseTessellation,
    Direction,
    OctobanTessellation,
    Tessellation,
    TriobanTessellation,
    index_1d,
    is_on_board_1d,
)
from sokoenginepy.io import CellOrientation


def triangle_points_down(position, board_width, board_height):
    return (
        TriobanTessellation().cell_orientation(position, board_width, board_height)
        == CellOrientation.TRIANGLE_DOWN
    )


def is_octagon(position, board_width, board_height):
    return (
        OctobanTessellation().cell_orientation(position, board_width, board_height)
        == CellOrientation.OCTAGON
    )


class SokobanTessellationAutogeneratedSpecMixin:
    class Describe_neighbor_position:
        def test_autogenerated_topLeft(self):
            width = 10
            height = 10
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(0,1, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_topRight(self):
            width = 10
            height = 10
            row = 0
            column = 9
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(8,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(9,1, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_bottomLeft(self):
            width = 10
            height = 10
            row = 9
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,9, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,8, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_bottomRight(self):
            width = 10
            height = 10
            row = 9
            column = 9
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(8,9, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(9,8, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_midleTop(self):
            width = 10
            height = 10
            row = 0
            column = 5
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(4,0, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(6,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(5,1, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_midleBottom(self):
            width = 10
            height = 10
            row = 9
            column = 5
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(4,9, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(6,9, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(5,8, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_midleLeft(self):
            width = 10
            height = 10
            row = 5
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,5, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,4, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(0,6, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_midleRight(self):
            width = 10
            height = 10
            row = 5
            column = 9
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(8,5, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(9,4, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(9,6, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_board_1x1(self):
            width = 1
            height = 1
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_board_2x1_left(self):
            width = 2
            height = 1
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_boarD_2x1_right(self):
            width = 2
            height = 1
            row = 0
            column = 1
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(0,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_1x2_top(self):
            width = 1
            height = 2
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(0,1, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)

        def test_autogenerated_1x2_bottom(self):
            width = 1
            height = 2
            row = 1
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.SOKOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_WEST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.NORTH_EAST, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.SOUTH_EAST, width, height)


class TriobanTessellationAutogeneratedSpecMixin:
    class Describe_neighbor_position:
        def test_autogenerated_topLeft(self):
            width = 5
            height = 5
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert triangle_points_down(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,0, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(1,0, width)

        def test_autogenerated_topRight_columnOdd(self):
            width = 4
            height = 4
            row = 0
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert column % 2 == 1 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,0, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,1, width)

        def test_autogenerated_topRight_columnEven(self):
            width = 5
            height = 5
            row = 0
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert column % 2 == 0 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomLeft_rowOdd(self):
            width = 4
            height = 4
            row = 3
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 1 and not triangle_points_down(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(1,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomLeft_rowEven(self):
            width = 5
            height = 5
            row = 4
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 0 and triangle_points_down(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,4, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(0,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(0,3, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(1,4, width)

        def test_autogenerated_bottomRight_rowEven_columnEven(self):
            width = 5
            height = 5
            row = 4
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 0 and column % 2 == 0 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(4,3, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,4, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowEven_columnOdd(self):
            width = 4
            height = 5
            row = 4
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 0 and column % 2 == 1 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowOdd_columnEven(self):
            width = 5
            height = 4
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 1 and column % 2 == 0 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowOdd_columnOdd(self):
            width = 4
            height = 4
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 1 and column % 2 == 1 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleTop_columnOdd(self):
            width = 10
            height = 10
            row = 0
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert column % 2 == 1 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,0, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,0, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,0, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,1, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,0, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,1, width)

        def test_autogenerated_midleTop_columnEven(self):
            width = 10
            height = 10
            row = 0
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert column % 2 == 0 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,0, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,0, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(5,0, width)

        def test_autogenerated_midleBottom_columnOdd_rowEven(self):
            width = 10
            height = 5
            row = 4
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert column % 2 == 1 and row % 2 == 0 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,4, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,4, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_middleBottom_columnEven_rowEven(self):
            width = 10
            height = 5
            row = 4
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert column % 2 == 0 and row % 2 == 0 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,4, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,4, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(4,3, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,4, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,3, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(5,4, width)

        def test_autogenerated_midleBottom_columnOdd_rowOdd(self):
            width = 10
            height = 4
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert column % 2 == 1 and row % 2 == 1 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(4,3, width)

        def test_autogenerated_middleBottom_columnEven_rowOdd(self):
            width = 10
            height = 4
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert column % 2 == 0 and row % 2 == 1 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(5,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleLeft_rowOdd(self):
            width = 10
            height = 10
            row = 3
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 1 and not triangle_points_down(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(0,4, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(0,4, width)

        def test_autogenerated_midleLeft_rowEven(self):
            width = 10
            height = 10
            row = 2
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 0 and triangle_points_down(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,2, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(0,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(0,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(1,2, width)

        def test_autogenerated_midleRight_rowOdd_columnOdd(self):
            width = 4
            height = 10
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 1 and column % 2 == 1 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleRight_rowEven_columnOdd(self):
            width = 4
            height = 10
            row = 2
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 0 and column % 2 == 1 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,3, width)

        def test_autogenerated_midleRight_rowOdd_columnEven(self):
            width = 5
            height = 10
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 1 and column % 2 == 0 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,3, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(4,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(4,4, width)

        def test_autogenerated_midleRight_rowEven_columnEven(self):
            width = 5
            height = 10
            row = 2
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 0 and column % 2 == 0 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(4,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midle_rowOdd_columnOdd(self):
            width = 10
            height = 10
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 1 and column % 2 == 1 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(4,3, width)

        def test_autogenerated_midle_rowOdd_columnEven(self):
            width = 10
            height = 10
            row = 3
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 1 and column % 2 == 0 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(3,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,4, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,3, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(2,4, width)

        def test_autogenerated_midle_rowEven_columnOdd(self):
            width = 10
            height = 10
            row = 2
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 0 and column % 2 == 1 and not triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,2, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,3, width)

        def test_autogenerated_midle_rowEven_columnEven(self):
            width = 10
            height = 10
            row = 2
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert row % 2 == 0 and column % 2 == 0 and triangle_points_down(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,2, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(3,2, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(1,2, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(2,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,2, width)

        def test_autogenerated_board_1x1(self):
            width = 1
            height = 1
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_board_2x1_left(self):
            width = 2
            height = 1
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,0, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(1,0, width)

        def test_autogenerated_board_2x1_right(self):
            width = 2
            height = 1
            row = 0
            column = 1
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(0,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(0,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_board_3x1_right(self):
            width = 3
            height = 1
            row = 0
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(1,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_1x2_top(self):
            width = 1
            height = 2
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_1x2_bottom(self):
            width = 1
            height = 2
            row = 1
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_1x3_bottom(self):
            width = 1
            height = 3
            row = 2
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.TRIOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(0,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(0,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)


class HexobanTessellationAutogeneratedSpecMixin:
    class Describe_neighbor_position:
        def test_autogenerated_topLeft(self):
            width = 5
            height = 5
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,0, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(0,1, width)

        def test_autogenerated_topRight_columnOdd(self):
            width = 4
            height = 4
            row = 0
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert column % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,1, width)

        def test_autogenerated_topRight_columnEven(self):
            width = 5
            height = 5
            row = 0
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert column % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(4,1, width)

        def test_autogenerated_bottomLeft_rowOdd(self):
            width = 4
            height = 4
            row = 3
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 1

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(0,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(1,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomLeft_rowEven(self):
            width = 5
            height = 5
            row = 4
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 0

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,4, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(0,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowEven_columnEven(self):
            width = 5
            height = 5
            row = 4
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 0 and column % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowEven_columnOdd(self):
            width = 4
            height = 5
            row = 4
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 0 and column % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowOdd_columnEven(self):
            width = 5
            height = 4
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 1 and column % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(4,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowOdd_columnOdd(self):
            width = 4
            height = 4
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 1 and column % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleTop_columnOdd(self):
            width = 10
            height = 10
            row = 0
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert column % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,0, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,0, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,1, width)

        def test_autogenerated_midleTop_columnEven(self):
            width = 10
            height = 10
            row = 0
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert column % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,0, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,0, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(4,1, width)

        def test_autogenerated_midleBottom_columnOdd_rowEven(self):
            width = 10
            height = 5
            row = 4
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert column % 2 == 1 and row % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,4, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,4, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_middleBottom_columnEven_rowEven(self):
            width = 10
            height = 5
            row = 4
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert column % 2 == 0 and row % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,4, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,4, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleBottom_columnOdd_rowOdd(self):
            width = 10
            height = 4
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert column % 2 == 1 and row % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_middleBottom_columnEven_rowOdd(self):
            width = 10
            height = 4
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert column % 2 == 0 and row % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(4,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(5,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleLeft_rowOdd(self):
            width = 10
            height = 10
            row = 3
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 1

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(0,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(0,4, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(1,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(1,4, width)

        def test_autogenerated_midleLeft_rowEven(self):
            width = 10
            height = 10
            row = 2
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 0

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,2, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(0,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(0,3, width)

        def test_autogenerated_midleRight_rowOdd_columnOdd(self):
            width = 4
            height = 10
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 1 and column % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleRight_rowEven_columnOdd(self):
            width = 4
            height = 10
            row = 2
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 0 and column % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,3, width)

        def test_autogenerated_midleRight_rowOdd_columnEven(self):
            width = 5
            height = 10
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 1 and column % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(4,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(4,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleRight_rowEven_columnEven(self):
            width = 5
            height = 10
            row = 2
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 0 and column % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(4,3, width)

        def test_autogenerated_midle_rowOdd_columnOdd(self):
            width = 10
            height = 10
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 1 and column % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,4, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(4,4, width)

        def test_autogenerated_midle_rowOdd_columnEven(self):
            width = 10
            height = 10
            row = 3
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 1 and column % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(3,3, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,4, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,4, width)

        def test_autogenerated_midle_rowEven_columnOdd(self):
            width = 10
            height = 10
            row = 2
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 0 and column % 2 == 1

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,2, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,3, width)

        def test_autogenerated_midle_rowEven_columnEven(self):
            width = 10
            height = 10
            row = 2
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert row % 2 == 0 and column % 2 == 0

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,2, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(3,2, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(1,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(2,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(2,3, width)

        def test_autogenerated_board_1x1(self):
            width = 1
            height = 1
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_board_2x1_left(self):
            width = 2
            height = 1
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,0, width)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_board_2x1_right(self):
            width = 2
            height = 1
            row = 0
            column = 1
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(0,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_board_3x1_right(self):
            width = 3
            height = 1
            row = 0
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_1x2_top(self):
            width = 1
            height = 2
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(0,1, width)

        def test_autogenerated_1x2_bottom(self):
            width = 1
            height = 2
            row = 1
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(0,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_1x3_bottom(self):
            width = 1
            height = 3
            row = 2
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.HEXOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.UP, width, height)
            with pytest.raises(ValueError):
                t.neighbor_position(index, Direction.DOWN, width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(0,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)


class OctobanTessellationAutogeneratedSpecMixin:
    class Describe_neighbor_position:
        def test_autogenerated_topLeft(self):
            width = 5
            height = 5
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert is_octagon(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(0,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(1,1, width)

        def test_autogenerated_topRight_columnOdd(self):
            width = 4
            height = 4
            row = 0
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert column % 2 == 1 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(3,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_topRight_columnEven(self):
            width = 5
            height = 5
            row = 0
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert column % 2 == 0 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(4,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomLeft_rowOdd(self):
            width = 4
            height = 4
            row = 3
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 1 and not is_octagon(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomLeft_rowEven(self):
            width = 5
            height = 5
            row = 4
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 0 and is_octagon(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,4, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(1,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowEven_columnEven(self):
            width = 5
            height = 5
            row = 4
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 0 and column % 2 == 0 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(4,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowEven_columnOdd(self):
            width = 4
            height = 5
            row = 4
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 0 and column % 2 == 1 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowOdd_columnEven(self):
            width = 5
            height = 4
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 1 and column % 2 == 0 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(4,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_bottomRight_rowOdd_columnOdd(self):
            width = 4
            height = 4
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 1 and column % 2 == 1 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleTop_columnOdd(self):
            width = 10
            height = 10
            row = 0
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert column % 2 == 1 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,0, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(3,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleTop_columnEven(self):
            width = 10
            height = 10
            row = 0
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert column % 2 == 0 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,0, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(4,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(5,1, width)

        def test_autogenerated_midleBottom_columnOdd_rowEven(self):
            width = 10
            height = 5
            row = 4
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert column % 2 == 1 and row % 2 == 0 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,4, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,4, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_middleBottom_columnEven_rowEven(self):
            width = 10
            height = 5
            row = 4
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert column % 2 == 0 and row % 2 == 0 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,4, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,4, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(4,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(5,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleBottom_columnOdd_rowOdd(self):
            width = 10
            height = 4
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert column % 2 == 1 and row % 2 == 1 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,3, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_middleBottom_columnEven_rowOdd(self):
            width = 10
            height = 4
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert column % 2 == 0 and row % 2 == 1 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(5,3, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(4,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleLeft_rowOdd(self):
            width = 10
            height = 10
            row = 3
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 1 and not is_octagon(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,2, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(0,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleLeft_rowEven(self):
            width = 10
            height = 10
            row = 2
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 0 and is_octagon(index, width, height)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,2, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,1, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(0,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(1,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(1,3, width)

        def test_autogenerated_midleRight_rowOdd_columnOdd(self):
            width = 4
            height = 10
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 1 and column % 2 == 1 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(3,4, width)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleRight_rowEven_columnOdd(self):
            width = 4
            height = 10
            row = 2
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 0 and column % 2 == 1 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(3,1, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleRight_rowOdd_columnEven(self):
            width = 5
            height = 10
            row = 3
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 1 and column % 2 == 0 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(4,2, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(4,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midleRight_rowEven_columnEven(self):
            width = 5
            height = 10
            row = 2
            column = 4
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 0 and column % 2 == 0 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(3,2, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(4,1, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(4,3, width)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(3,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midle_rowOdd_columnOdd(self):
            width = 10
            height = 10
            row = 3
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 1 and column % 2 == 1 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,3, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(3,4, width)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(2,4, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(4,2, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(4,4, width)

        def test_autogenerated_midle_rowOdd_columnEven(self):
            width = 10
            height = 10
            row = 3
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 1 and column % 2 == 0 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(3,3, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(2,4, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midle_rowEven_columnOdd(self):
            width = 10
            height = 10
            row = 2
            column = 3
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 0 and column % 2 == 1 and not is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(2,2, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(4,2, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(3,1, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(3,3, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_midle_rowEven_columnEven(self):
            width = 10
            height = 10
            row = 2
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert row % 2 == 0 and column % 2 == 0 and is_octagon(index, width, height)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,2, width)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(3,2, width)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(2,1, width)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(2,3, width)
            assert t.neighbor_position(index, Direction.NORTH_WEST, width, height) == index_1d(1,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_WEST, width, height) == index_1d(1,3, width)
            assert t.neighbor_position(index, Direction.NORTH_EAST, width, height) == index_1d(3,1, width)
            assert t.neighbor_position(index, Direction.SOUTH_EAST, width, height) == index_1d(3,3, width)

        def test_autogenerated_board_1x1(self):
            width = 1
            height = 1
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_board_2x1_left(self):
            width = 2
            height = 1
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert t.neighbor_position(index, Direction.RIGHT, width, height) == index_1d(1,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_board_2x1_right(self):
            width = 2
            height = 1
            row = 0
            column = 1
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(0,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_board_3x1_right(self):
            width = 3
            height = 1
            row = 0
            column = 2
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert t.neighbor_position(index, Direction.LEFT, width, height) == index_1d(1,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_1x2_top(self):
            width = 1
            height = 2
            row = 0
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.UP, width, height), width, height)
            assert t.neighbor_position(index, Direction.DOWN, width, height) == index_1d(0,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_1x2_bottom(self):
            width = 1
            height = 2
            row = 1
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,0, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)

        def test_autogenerated_1x3_bottom(self):
            width = 1
            height = 3
            row = 2
            column = 0
            index = index_1d(column, row, width)
            t = BaseTessellation.instance(Tessellation.OCTOBAN)

            assert not is_on_board_1d(t.neighbor_position(index, Direction.LEFT, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.RIGHT, width, height), width, height)
            assert t.neighbor_position(index, Direction.UP, width, height) == index_1d(0,1, width)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.DOWN, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_WEST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.NORTH_EAST, width, height), width, height)
            assert not is_on_board_1d(t.neighbor_position(index, Direction.SOUTH_EAST, width, height), width, height)





from dataclasses import dataclass
from pathlib import Path

import pytest
from faker import Faker

from sokoenginepy.game import PusherStep

_SELF_DIR = Path(__file__).parent.absolute().resolve()


fake = Faker()


@dataclass
class BoardData:
    board: str
    width: int
    height: int


@pytest.fixture(scope="session")
def is_using_native():
    return PusherStep.__module__ == "sokoenginepyext.game"


@pytest.fixture(scope="session")
def resources_root():
    return _SELF_DIR



"""
Module that handles Sokoban I/O and text processing.
"""

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .collection import Collection
    from .hexoban import HexobanPuzzle, HexobanSnapshot
    from .octoban import OctobanPuzzle, OctobanSnapshot
    from .puzzle import CellOrientation, Puzzle
    from .rle import Rle
    from .snapshot import Snapshot
    from .sokoban import SokobanPuzzle, SokobanSnapshot
    from .trioban import TriobanPuzzle, TriobanSnapshot

else:
    try:
        from sokoenginepyext.io import (
            CellOrientation,
            Collection,
            HexobanPuzzle,
            HexobanSnapshot,
            OctobanPuzzle,
            OctobanSnapshot,
            Puzzle,
            Rle,
            Snapshot,
            SokobanPuzzle,
            SokobanSnapshot,
            TriobanPuzzle,
            TriobanSnapshot,
        )

    except ImportError:
        from .collection import Collection
        from .hexoban import HexobanPuzzle, HexobanSnapshot
        from .octoban import OctobanPuzzle, OctobanSnapshot
        from .puzzle import CellOrientation, Puzzle
        from .rle import Rle
        from .snapshot import Snapshot
        from .sokoban import SokobanPuzzle, SokobanSnapshot
        from .trioban import TriobanPuzzle, TriobanSnapshot

from .utilities import is_blank



import glob

import inspect
import os
from inspect import getsourcefile

import pytest

from sokoenginepy import game, io


def _get_filtered_members(mdl):
    return {
        name
        for name, obj in inspect.getmembers(mdl)
        if not inspect.ismodule(obj) and not name.startswith("__")
    }


@pytest.fixture
def io_members():
    return {
        "CellOrientation",
        "Collection",
        "HexobanPuzzle",
        "HexobanSnapshot",
        "OctobanPuzzle",
        "OctobanSnapshot",
        "Puzzle",
        "Rle",
        "Snapshot",
        "SokobanPuzzle",
        "SokobanSnapshot",
        "TriobanPuzzle",
        "TriobanSnapshot",
    }


@pytest.fixture
def io_members_py_only():
    # Stuff that is always used from Python implementation, even when native extension
    # is built and loaded
    return {
        "is_blank",
    }


@pytest.fixture
def game_members():
    return {
        "BaseTessellation",
        "BoardCell",
        "BoardGraph",
        "BoardManager",
        "BoardState",
        "BoxGoalSwitchError",
        "CellAlreadyOccupiedError",
        "Direction",
        "GraphType",
        "HashedBoardManager",
        "HexobanTessellation",
        "IllegalMoveError",
        "Mover",
        "NonPlayableBoardError",
        "OctobanTessellation",
        "PusherStep",
        "SokobanPlus",
        "SokobanPlusDataError",
        "SokobanTessellation",
        "SolvingMode",
        "Tessellation",
        "TriobanTessellation",
        "Config",
    }


@pytest.fixture
def game_members_native_only():
    return {
        # Makes no sense in Python layer because board size is not limited (yet) in
        # Python
        "BoardSizeExceededError",
        # Mapped to Python KeyError
        "ExtKeyError",
    }


@pytest.fixture
def game_members_py_only():
    # Stuff that is always used from Python implementation, even when native extension
    # is built and loaded
    return {
        "COLUMN",
        "JumpCommand",
        "MoveCommand",
        "ROW",
        "SelectPusherCommand",
        "X",
        "Y",
        "index_1d",
        "is_on_board_1d",
        "is_on_board_2d",
    }


class DescribeNativeCppExtension:
    def it_is_build_and_loaded_if_environment_required_it(self, is_using_native):
        building_on_travis = os.environ.get("TRAVIS", None)

        if building_on_travis:
            job_name = os.environ.get("TRAVIS_JOB_NAME", None)

            if job_name and "native extension" in job_name:
                assert os.environ.get("SOKOENGINEPYEXT_BUILD", None) == "true"
                assert is_using_native

            else:
                assert os.environ.get("SOKOENGINEPYEXT_BUILD", None) == "false"
                assert not is_using_native

        else:
            src_dir = os.path.abspath(
                os.path.join(getsourcefile(lambda: 0), "..", "..", "src")
            )
            ext_built = glob.glob(os.path.join(src_dir, "sokoenginepyext*.so"))

            if ext_built:
                assert is_using_native
            else:
                assert not is_using_native

    def it_is_correctly_imported_if_built(
        self, is_using_native, io_members, game_members, game_members_native_only
    ):
        if not is_using_native:
            return

        for member in io_members:
            assert getattr(io, member).__module__.startswith(
                "sokoenginepyext.io"
            ), f"{member} should've been imported from C++ implementation!"

        for member in game_members:
            assert getattr(game, member).__module__.startswith(
                "sokoenginepyext.game"
            ), f"{member} should've been imported from C++ implementation!"

        from sokoenginepyext import game as native_game

        for member in game_members_native_only:
            assert hasattr(
                native_game, member
            ), f"C++ extension should export {member}. Did you change sokoenginepyext code?"

    def it_is_not_imported_if_not_built(
        self, is_using_native, io_members, game_members
    ):
        if is_using_native:
            return

        for member in io_members:
            assert getattr(io, member).__module__.startswith(
                "sokoenginepy.io"
            ), f"{member} should've been imported from Python implementation!"

        for member in game_members:
            assert getattr(game, member).__module__.startswith(
                "sokoenginepy.game"
            ), f"{member} should've been imported from Python implementation!"

    def test_some_things_are_always_used_from_python_implementation(
        self, game_members_py_only, io_members_py_only
    ):
        for member in game_members_py_only:
            assert hasattr(
                game, member
            ), f"'sokoenginepy.game' package should have {member}!"

            module_name = getattr(game, member).__module__
            assert (
                module_name.startswith("sokoenginepy.game") or module_name == "typing"
            ), (
                f"{member} should be imported from Python implementation even if "
                "C++ extension is used!"
            )

        for member in io_members_py_only:
            assert hasattr(
                io, member
            ), f"'sokoenginepy.io' package should have {member}!"

            module_name = getattr(io, member).__module__
            assert (
                module_name.startswith("sokoenginepy.io") or module_name == "typing"
            ), (
                f"{member} should be imported from Python implementation even if "
                "C++ extension is used!"
            )



from __future__ import annotations

from typing import Dict, Final, List, Optional

from .config import Config


class SokobanPlusDataError(ValueError):
    pass


class SokobanPlus:
    """
    Manages Sokoban+ data for game board.

    **Sokoban+ rules**

    In this variant of game rules, each box and each goal on board get number tag
    (color). Game objective changes slightly: board is considered solved only when
    each goal is filled with box of the same tag. So, for example goal tagged with
    number 9 must be filled with any box tagged with number 9.

    Multiple boxes and goals may share same plus id, but the number of boxes with one
    plus id must be equal to number of goals with that same plus id. There is also
    default plus id that represents non tagged boxes and goals.

    Sokoban+ ids for given board are defined by two strings called goalorder and
    boxorder. For example, boxorder "13 24 3 122 1" would give plus_id = 13 to box id
    = 1, plus_id = 24 to box ID = 2, etc...

    **Valid Sokoban+ id sequences**

    Boxorder and goalorder must define ids for equal number of boxes and goals. This
    means that in case of boxorder assigning plus id "42" to two boxes, goalorder
    must also contain number 42 twice.

    Sokoban+ data parser accepts any positive integer as plus id.
    """

    LEGACY_DEFAULT_PLUS_ID: Final[int] = 99

    #: Sokoban+ ID for pieces that don't have one or when Sokoban+ is disabled.
    #:
    #: Original Sokoban+ implementation used number 99 for default plus ID. As
    #: there can be more than 99 boxes on board, sokoenginepy changes this
    #: detail and uses :const:`DEFAULT_PLUS_ID` as default plus ID. When loading
    #: older puzzles with Sokoban+, legacy default value is converted
    #: transparently.
    DEFAULT_PLUS_ID: Final[int] = 0

    def __init__(
        self,
        pieces_count: int,
        boxorder: Optional[str] = None,
        goalorder: Optional[str] = None,
    ):
        """
        Args:
            boxorder: Space separated integers describing Sokoban+ IDs for boxes
            goalorder: Space separated integers describing Sokoban+ IDs for goals
            pieces_count: Total count of boxes/goals on board
        """
        self._is_enabled = False
        self._is_validated = False
        self._errors: List[str] = []

        self._pieces_count = pieces_count
        self._box_plus_ids: Dict[int, int] = {}
        self._goal_plus_ids: Dict[int, int] = {}
        self._boxorder: str = boxorder or ""
        self._goalorder: str = goalorder or ""

    @classmethod
    def is_valid_plus_id(cls, plus_id: int) -> bool:
        return isinstance(plus_id, int) and plus_id >= cls.DEFAULT_PLUS_ID

    @property
    def pieces_count(self) -> int:
        return self._pieces_count

    @pieces_count.setter
    def pieces_count(self, rv):
        if rv != self._pieces_count:
            self.is_enabled = False
            self._is_validated = False
            self._pieces_count = int(rv)

    @property
    def boxorder(self) -> str:
        if self.is_enabled and self.is_valid:
            return self._rstrip_default_plus_ids(
                " ".join(str(i) for i in self._box_plus_ids.values())
            )
        else:
            return self._boxorder

    @boxorder.setter
    def boxorder(self, rv):
        if rv != self._boxorder:
            self.is_enabled = False
            self._is_validated = False
            self._boxorder = rv or ""

    @property
    def goalorder(self) -> str:
        if self.is_enabled and self.is_valid:
            return self._rstrip_default_plus_ids(
                " ".join(str(i) for i in self._goal_plus_ids.values())
            )
        else:
            return self._goalorder

    @goalorder.setter
    def goalorder(self, rv):
        if rv != self._goalorder:
            self.is_enabled = False
            self._is_validated = False
            self._goalorder = rv or ""

    @property
    def is_valid(self) -> bool:
        if self._is_validated:
            return not self._errors

        self._errors = []
        try:
            self._box_plus_ids = self._parse_and_clean_ids_string(self._boxorder)
            self._goal_plus_ids = self._parse_and_clean_ids_string(self._goalorder)
        except ValueError as exc:
            self._errors.append(str(exc))

        self._validate_plus_ids(self._box_plus_ids)
        self._validate_plus_ids(self._goal_plus_ids)
        self._validate_piece_count()
        self._validate_ids_counts()
        self._validate_id_sets_equality()
        self._is_validated = True

        return not self._errors

    @property
    def is_enabled(self) -> bool:
        return self._is_enabled

    @property
    def errors(self) -> List[str]:
        return self._errors

    @is_enabled.setter
    def is_enabled(self, value):
        """
        Raises:
            :exc:`SokobanPlusDataError`: Trying to enable invalid Sokoban+
        """
        if value:
            if not self.is_valid:
                raise SokobanPlusDataError(self._errors)
        self._is_enabled = value

    @property
    def is_validated(self):
        return self._is_validated

    def box_plus_id(self, for_box_id: int) -> int:
        """
        Get Sokoban+ ID for box.

        Returns:
            If Sokoban+ is enabled returns Sokoban+ ID of a box. If not, returns
            :const:`DEFAULT_PLUS_ID`

        Raises:
            :exc:`KeyError`: No box with ID ``for_box_id``, but only if i Sokoban+ is
                enabled
        """
        try:
            return self._get_plus_id(for_box_id, from_where=self._box_plus_ids)
        except KeyError:
            raise KeyError("No box with ID: {0}".format(for_box_id))

    def goal_plus_id(self, for_goal_id: int) -> int:
        """
        Get Sokoban+ ID for goal.

        Returns:
            If Sokoban+ is enabled returns Sokoban+ ID of a goal. If not,
            returns :const:`DEFAULT_PLUS_ID`

        Raises:
            :exc:`KeyError`: No goal with ID ``for_goal_id``, but only if Sokoban+ is
                enabled
        """
        try:
            return self._get_plus_id(for_goal_id, from_where=self._goal_plus_ids)
        except KeyError:
            raise KeyError("No goal with ID: {0}".format(for_goal_id))

    def _rstrip_default_plus_ids(self, plus_ids_str):
        # TODO: Might not work correctly for "3 5 4 6 2 19" or "3 5 4 6 2 10"
        if self.pieces_count < self.LEGACY_DEFAULT_PLUS_ID:
            return plus_ids_str.rstrip(
                str(self.DEFAULT_PLUS_ID) + " " + str(self.LEGACY_DEFAULT_PLUS_ID)
            )
        else:
            return plus_ids_str.rstrip(str(self.DEFAULT_PLUS_ID) + " ")

    def _get_plus_id(self, for_id: int, from_where: Dict[int, int]):
        if not self.is_enabled:
            return self.DEFAULT_PLUS_ID
        else:
            return from_where[for_id]

    def _parse_and_clean_ids_string(self, plus_ids_str) -> Dict[int, int]:
        """
        Safely replaces legacy default plus ids with default ones.

        Returns:
            dict: dict that maps piece IDs to piece Sokoban+ IDs
        """

        def convert_or_raise(id_str):
            try:
                return int(id_str)
            except ValueError:
                raise SokobanPlusDataError(
                    "Can't parse Sokoban+ string! Illegal characters found. "
                    "Only digits and spaces allowed."
                )

        trimmed = [
            convert_or_raise(id_str)
            for id_str in self._rstrip_default_plus_ids(plus_ids_str).split()
        ]

        cleaned = [
            self.DEFAULT_PLUS_ID
            if (
                i == self.LEGACY_DEFAULT_PLUS_ID
                and self.pieces_count < self.LEGACY_DEFAULT_PLUS_ID
            )
            else i
            for i in trimmed
        ]

        expanded = cleaned + [self.DEFAULT_PLUS_ID] * (self.pieces_count - len(cleaned))

        retv = dict()
        for index, plus_id in enumerate(expanded):
            retv[Config.DEFAULT_PIECE_ID + index] = plus_id

        return retv

    def _validate_plus_ids(self, ids: Dict[int, int]):
        if ids:
            for i in ids.values():
                if not self.is_valid_plus_id(i):
                    self._errors.append(f"Invalid Sokoban+ ID: {i}")

    def _validate_piece_count(self):
        if self.pieces_count < 0:
            self._errors.append("Sokoban+ can't be applied to zero pieces count.")

    def _validate_ids_counts(self):
        if self._box_plus_ids and len(self._box_plus_ids) != self.pieces_count:
            self._errors.append(
                "Sokoban+ boxorder data doesn't contain same amount of IDs "
                + "as there are pieces on board! (pieces_count: {0})".format(
                    self.pieces_count
                )
            )

        if self._goal_plus_ids and len(self._goal_plus_ids) != self.pieces_count:
            self._errors.append(
                "Sokoban+ goalorder data doesn't contain same amount of IDs "
                + "as there are pieces on board! (pieces_count: {0})".format(
                    self.pieces_count
                )
            )

    def _validate_id_sets_equality(self):
        if self._box_plus_ids:
            boxes = set(
                pid
                for pid in self._box_plus_ids.values()
                if pid != self.DEFAULT_PLUS_ID
            )
        else:
            boxes = set()

        if self._goal_plus_ids:
            goals = set(
                pid
                for pid in self._goal_plus_ids.values()
                if pid != self.DEFAULT_PLUS_ID
            )
        else:
            goals = set()

        if boxes != goals:
            self._errors.append(
                "Sokoban+ data doesn't define equal sets of IDs for "
                + "boxes and goals"
            )

    def __repr__(self):
        return (
            "SokobanPlus("
            + "pieces_count={0}, boxorder='{1}', goalorder='{2}')".format(
                self.pieces_count, self.boxorder, self.goalorder
            )
        )

    def __str__(self):
        return (
            "SokobanPlus("
            + "pieces_count={0}, boxorder='{1}', goalorder='{2}')".format(
                self.pieces_count, self.boxorder, self.goalorder
            )
        )
